package compiler

import (
	"bufio"
	"fmt"
	"io"
	"os"

	"git.akyoto.dev/cli/q/src/asm"
	"git.akyoto.dev/cli/q/src/config"
	"git.akyoto.dev/cli/q/src/core"
	"git.akyoto.dev/cli/q/src/dll"
	"git.akyoto.dev/cli/q/src/elf"
	"git.akyoto.dev/cli/q/src/macho"
	"git.akyoto.dev/cli/q/src/pe"
	"git.akyoto.dev/cli/q/src/x86"
	"git.akyoto.dev/go/color/ansi"
)

// Result contains all the compiled functions in a build.
type Result struct {
	Main             *core.Function
	Functions        map[string]*core.Function
	InstructionCount int
	DataCount        int
	Code             []byte
	Data             []byte
	DLLs             dll.List
}

// finalize generates the final machine code.
func (r *Result) finalize() {
	// This will be the entry point of the executable.
	// The only job of the entry function is to call `main` and exit cleanly.
	// The reason we call `main` instead of using `main` itself is to place
	// a return address on the stack, which allows return statements in `main`.
	final := asm.Assembler{
		Instructions: make([]asm.Instruction, 0, r.InstructionCount+8),
		Data:         make(map[string][]byte, r.DataCount),
	}

	final.Call("main.main")

	switch config.TargetOS {
	case config.Linux:
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[0], LinuxExit)
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[1], 0)
		final.Syscall()
	case config.Mac:
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[0], MacExit)
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[1], 0)
		final.Syscall()
	case config.Windows:
		final.RegisterNumber(asm.MOVE, x86.WindowsInputRegisters[0], 0)
		final.DLLCall("kernel32.ExitProcess")
	}

	r.DLLs = dll.List{
		{Name: "kernel32", Functions: []string{"ExitProcess"}},
	}

	// This will place the main function immediately after the entry point
	// and also add everything the main function calls recursively.
	r.eachFunction(r.Main, map[*core.Function]bool{}, func(f *core.Function) {
		final.Merge(f.Assembler)

		for _, library := range f.DLLs {
			for _, fn := range library.Functions {
				r.DLLs = r.DLLs.Append(library.Name, fn)
			}
		}
	})

	final.Label(asm.LABEL, "_crash")

	switch config.TargetOS {
	case config.Linux:
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[0], LinuxExit)
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[1], 1)
		final.Syscall()
	case config.Mac:
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[0], MacExit)
		final.RegisterNumber(asm.MOVE, x86.SyscallInputRegisters[1], 1)
		final.Syscall()
	case config.Windows:
		final.RegisterNumber(asm.MOVE, x86.WindowsInputRegisters[0], 1)
		final.DLLCall("kernel32.ExitProcess")
	}

	r.Code, r.Data = final.Finalize(r.DLLs)
}

// eachFunction recursively finds all the calls to external functions.
// It avoids calling the same function twice with the help of a hashmap.
func (r *Result) eachFunction(caller *core.Function, traversed map[*core.Function]bool, call func(*core.Function)) {
	call(caller)
	traversed[caller] = true

	for _, x := range caller.Assembler.Instructions {
		if x.Mnemonic != asm.CALL {
			continue
		}

		name := x.Data.(*asm.Label).Name
		callee, exists := r.Functions[name]

		if !exists {
			continue
		}

		if traversed[callee] {
			continue
		}

		r.eachFunction(callee, traversed, call)
	}
}

// PrintInstructions prints out the generated instructions.
func (r *Result) PrintInstructions() {
	r.eachFunction(r.Main, map[*core.Function]bool{}, func(f *core.Function) {
		f.PrintInstructions()
	})
}

// PrintStatistics shows the statistics.
func (r *Result) PrintStatistics() {
	ansi.Dim.Println("╭──────────────────────────────────────────────────────────────────────────────╮")
	ansi.Dim.Printf("│ %-44s%-32s │\n", "Code:", fmt.Sprintf("%d bytes", len(r.Code)))
	ansi.Dim.Printf("│ %-44s%-32s │\n", "Data:", fmt.Sprintf("%d bytes", len(r.Data)))
	ansi.Dim.Println("╰──────────────────────────────────────────────────────────────────────────────╯")
}

// Write writes the executable to the given writer.
func (r *Result) Write(writer io.Writer) error {
	return write(writer, r.Code, r.Data, r.DLLs)
}

// Write writes an executable file to disk.
func (r *Result) WriteFile(path string) error {
	file, err := os.Create(path)

	if err != nil {
		return err
	}

	err = r.Write(file)

	if err != nil {
		file.Close()
		return err
	}

	err = file.Close()

	if err != nil {
		return err
	}

	return os.Chmod(path, 0755)
}

// write writes an executable file to the given writer.
func write(writer io.Writer, code []byte, data []byte, dlls dll.List) error {
	buffer := bufio.NewWriter(writer)

	switch config.TargetOS {
	case config.Linux:
		elf.Write(buffer, code, data)
	case config.Mac:
		macho.Write(buffer, code, data)
	case config.Windows:
		pe.Write(buffer, code, data, dlls)
	}

	return buffer.Flush()
}
